# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import copy

import mock

from vmware_nsxlib.tests.unit.v3 import nsxlib_testcase
from vmware_nsxlib.tests.unit.v3.policy import policy_testcase
from vmware_nsxlib.v3 import policy
from vmware_nsxlib.v3.policy import constants
from vmware_nsxlib.v3.policy import transaction as trans


class TestPolicyTransaction(policy_testcase.TestPolicyApi):

    def setUp(self):

        super(TestPolicyTransaction, self).setUp()

        nsxlib_config = nsxlib_testcase.get_default_nsxlib_config()
        # Mock the nsx-lib for the passthrough api
        with mock.patch('vmware_nsxlib.v3.NsxLib.get_version',
                        return_value='2.5.0'):

            self.policy_lib = policy.NsxPolicyLib(nsxlib_config)
        self.policy_api = self.policy_lib.policy_api
        self.policy_api.client = self.client

    def assert_infra_patch_call(self, body):
        self.assert_json_call('PATCH', self.client, 'infra',
                              data=body, headers=mock.ANY)

    def test_domains_only(self):

        tags = [{'scope': 'color', 'tag': 'green'}]
        d1 = {'resource_type': 'Domain', 'id': 'domain1',
              'display_name': 'd1', 'description': 'first domain',
              'tags': tags}

        d2 = {'resource_type': 'Domain', 'id': 'domain2',
              'display_name': 'd2', 'description': 'no tags',
              'tags': None}
        with trans.NsxPolicyTransaction():

            for d in (d1, d2):
                self.policy_lib.domain.create_or_overwrite(
                    d['display_name'],
                    d['id'],
                    d['description'],
                    tags=d['tags'] if 'tags' in d else None)

        expected_body = {'resource_type': 'Infra',
                         'children': [{'resource_type': 'ChildDomain',
                                       'Domain': d1},
                                      {'resource_type': 'ChildDomain',
                                       'Domain': d2}]}

        self.assert_infra_patch_call(expected_body)

    def test_domains_and_groups(self):

        tags = [{'scope': 'color', 'tag': 'green'}]
        g1 = {'resource_type': 'Group', 'id': 'group1',
              'display_name': 'g1',
              'description': 'first group',
              'tags': None, 'expression': []}
        g2 = {'resource_type': 'Group', 'id': 'group2',
              'description': 'second group',
              'display_name': 'g2',
              'tags': tags, 'expression': []}
        g3 = {'resource_type': 'Group', 'id': 'group3',
              'display_name': 'g3',
              'description': 'third group',
              'tags': None, 'expression': []}
        d1 = {'resource_type': 'Domain', 'id': 'domain1',
              'display_name': 'd1', 'description': 'first domain',
              'tags': tags}

        d2 = {'resource_type': 'Domain', 'id': 'domain2',
              'display_name': 'd2', 'description': 'no tags',
              'tags': None}

        with trans.NsxPolicyTransaction():

            for d in (d1, d2):
                self.policy_lib.domain.create_or_overwrite(
                    d['display_name'],
                    d['id'],
                    d['description'],
                    tags=d['tags'] if 'tags' in d else None)

                d['children'] = []

                for g in (g1, g2, g3):
                    self.policy_lib.group.create_or_overwrite(
                        g['display_name'],
                        d['id'],
                        g['id'],
                        g['description'],
                        tags=g['tags'] if 'tags' in g else None)

                    d['children'].append({'resource_type': 'ChildGroup',
                                          'Group': g})

        expected_body = {'resource_type': 'Infra',
                         'children': [{'resource_type': 'ChildDomain',
                                       'Domain': d1},
                                      {'resource_type': 'ChildDomain',
                                       'Domain': d2}]}

        self.assert_infra_patch_call(expected_body)

    def test_ip_address_pool_and_block_subnets(self):

        pool = {'id': 'pool1',
                'resource_type': 'IpAddressPool',
                'display_name': 'pool1',
                'children': []}

        ip_block_id = 'block1'
        subnet1 = {'id': 'subnet1',
                   'resource_type': 'IpAddressPoolBlockSubnet',
                   'ip_block_path': '/infra/ip-blocks/%s' % ip_block_id,
                   'size': 8}

        subnet2 = {'id': 'subnet2',
                   'resource_type': 'IpAddressPoolBlockSubnet',
                   'ip_block_path': '/infra/ip-blocks/%s' % ip_block_id,
                   'size': 4}

        with trans.NsxPolicyTransaction():
            self.policy_lib.ip_pool.create_or_overwrite(
                pool['display_name'],
                ip_pool_id=pool['id'])

            for s in (subnet1, subnet2):
                self.policy_lib.ip_pool.allocate_block_subnet(
                    ip_pool_id=pool['id'],
                    ip_block_id=ip_block_id,
                    ip_subnet_id=s['id'],
                    size=s['size'])

                pool['children'].append(
                    {'resource_type': 'ChildIpAddressPoolSubnet',
                     'IpAddressPoolSubnet': s})

        expected_body = {'resource_type': 'Infra',
                         'children': [{'resource_type': 'ChildIpAddressPool',
                                       'IpAddressPool': pool}]}

        self.assert_infra_patch_call(expected_body)

    def test_groups_only(self):

        g1 = {'resource_type': 'Group', 'id': 'group1',
              'display_name': 'g1',
              'description': 'first group', 'expression': []}
        g2 = {'resource_type': 'Group', 'id': 'group2',
              'description': 'second group',
              'display_name': 'g2', 'expression': []}
        d1 = {'resource_type': 'Domain', 'id': 'domain1'}

        d2 = {'resource_type': 'Domain', 'id': 'domain2'}

        with trans.NsxPolicyTransaction():

            for d in (d1, d2):
                d['children'] = []

                for g in (g1, g2):
                    self.policy_lib.group.create_or_overwrite(
                        g['display_name'],
                        d['id'],
                        g['id'],
                        g['description'])

                    d['children'].append({'resource_type': 'ChildGroup',
                                          'Group': g})

        expected_body = {'resource_type': 'Infra',
                         'children': [{'resource_type': 'ChildDomain',
                                       'Domain': d1},
                                      {'resource_type': 'ChildDomain',
                                       'Domain': d2}]}

        self.assert_infra_patch_call(expected_body)

    def test_segment_ports(self):

        port1 = {'id': 'port_on_seg1',
                 'resource_type': 'SegmentPort',
                 'display_name': 'port_on_seg1',
                 'attachment': {'type': 'VIF',
                                'app_id': 'app1',
                                'traffic_tag': 5}
                 }

        port2 = {'id': 'port1_on_seg2',
                 'resource_type': 'SegmentPort',
                 'display_name': 'port_on_seg2',
                 'attachment': {'type': 'CHILD',
                                'app_id': 'app2',
                                'traffic_tag': None}
                 }

        seg1 = {'id': 'seg1',
                'resource_type': 'Segment',
                'children': [{'resource_type': 'ChildSegmentPort',
                              'SegmentPort': port1}]}
        seg2 = {'id': 'seg2',
                'resource_type': 'Segment',
                'children': [{'resource_type': 'ChildSegmentPort',
                              'SegmentPort': port2}]}

        with trans.NsxPolicyTransaction():

            self.policy_lib.segment_port.create_or_overwrite(
                port1['display_name'],
                seg1['id'],
                port1['id'],
                attachment_type=port1['attachment']['type'],
                app_id=port1['attachment']['app_id'],
                traffic_tag=port1['attachment']['traffic_tag'])

            self.policy_lib.segment_port.create_or_overwrite(
                port2['display_name'],
                seg2['id'],
                port2['id'],
                attachment_type=port2['attachment']['type'],
                app_id=port2['attachment']['app_id'],
                traffic_tag=port2['attachment']['traffic_tag'])

        expected_body = {'resource_type': 'Infra',
                         'children': [{'resource_type': 'ChildSegment',
                                       'Segment': seg1},
                                      {'resource_type': 'ChildSegment',
                                       'Segment': seg2}]}

        self.assert_infra_patch_call(expected_body)

    def test_tier1_nat_rules_create(self):
        tier1_id = 'tier1-1'
        nat_rule_id1 = 'nat1'
        nat_rule_id2 = 'nat2'

        nat_rule1 = {"action": constants.NAT_ACTION_SNAT,
                     "display_name": "snat rule",
                     "id": nat_rule_id1,
                     "resource_type": "PolicyNatRule"}
        nat_rule2 = {"action": constants.NAT_ACTION_DNAT,
                     "display_name": "dnat rule",
                     "id": nat_rule_id2,
                     "resource_type": "PolicyNatRule"}

        policy_nat = {"id": "USER",
                      "resource_type": "PolicyNat",
                      "children": [
                          {"PolicyNatRule": nat_rule1,
                           "resource_type": "ChildPolicyNatRule"},
                          {"PolicyNatRule": nat_rule2,
                           "resource_type": "ChildPolicyNatRule"}]}
        tier1_dict = {"id": tier1_id,
                      "resource_type": "Tier1",
                      "children": [{"PolicyNat": policy_nat,
                                    "resource_type": "ChildPolicyNat"}]}

        with trans.NsxPolicyTransaction():
            self.policy_lib.tier1_nat_rule.create_or_overwrite(
                'snat rule',
                tier1_id,
                nat_rule_id=nat_rule_id1,
                action=constants.NAT_ACTION_SNAT)

            self.policy_lib.tier1_nat_rule.create_or_overwrite(
                'dnat rule',
                tier1_id,
                nat_rule_id=nat_rule_id2,
                action=constants.NAT_ACTION_DNAT)

        expected_body = {"resource_type": "Infra",
                         "children": [{"Tier1": tier1_dict,
                                       "resource_type": "ChildTier1"}]}

        self.assert_infra_patch_call(expected_body)

    def test_tier1_nat_rules_delete(self):
        tier1_id = 'tier1-1'
        nat_rule_id1 = 'nat1'
        nat_rule_id2 = 'nat2'

        nat_rule1 = {"action": constants.NAT_ACTION_DNAT,
                     "id": nat_rule_id1,
                     "resource_type": "PolicyNatRule"}
        nat_rule2 = {"action": constants.NAT_ACTION_DNAT,
                     "id": nat_rule_id2,
                     "resource_type": "PolicyNatRule"}

        policy_nat = {"id": "USER",
                      "resource_type": "PolicyNat",
                      "children": [
                          {"PolicyNatRule": nat_rule1,
                           "marked_for_delete": True,
                           "resource_type": "ChildPolicyNatRule"},
                          {"PolicyNatRule": nat_rule2,
                           "marked_for_delete": True,
                           "resource_type": "ChildPolicyNatRule"}]}
        tier1_dict = {"id": tier1_id,
                      "resource_type": "Tier1",
                      "children": [{"PolicyNat": policy_nat,
                                    "resource_type": "ChildPolicyNat"}]}

        with trans.NsxPolicyTransaction():
            self.policy_lib.tier1_nat_rule.delete(
                tier1_id,
                nat_rule_id=nat_rule_id1)

            self.policy_lib.tier1_nat_rule.delete(
                tier1_id,
                nat_rule_id=nat_rule_id2)

        expected_body = {"resource_type": "Infra",
                         "children": [{"Tier1": tier1_dict,
                                       "resource_type": "ChildTier1"}]}

        self.assert_infra_patch_call(expected_body)

    def test_creating_security_policy_and_dfw_rules(self):
        dfw_rule = {'id': 'rule_id1', 'action': 'ALLOW',
                    'display_name': 'rule1', 'description': None,
                    'direction': 'IN_OUT', 'ip_protocol': 'IPV4_IPV6',
                    'logged': False, 'destination_groups': ['destination_url'],
                    'source_groups': ['src_url'], 'resource_type': 'Rule',
                    'scope': None, 'sequence_number': None, 'tag': None,
                    'services': ['ANY']}
        security_policy = {'id': 'security_policy_id1',
                           'display_name': 'security_policy',
                           'category': 'Application',
                           'resource_type': 'SecurityPolicy'}
        domain = {'resource_type': 'Domain', 'id': 'domain1'}
        domain_id = domain['id']
        map_id = security_policy['id']
        dfw_rule_entries = [self.policy_lib.comm_map.build_entry(
            name=dfw_rule['display_name'],
            domain_id=domain_id,
            map_id=map_id,
            entry_id=dfw_rule['id'],
            source_groups=dfw_rule['source_groups'],
            dest_groups=dfw_rule['destination_groups']
        )]
        with trans.NsxPolicyTransaction():
            self.policy_lib.comm_map.create_with_entries(
                name=security_policy['display_name'],
                domain_id=domain_id,
                map_id=map_id,
                entries=dfw_rule_entries
            )

        def get_group_path(group_id, domain_id):
            return '/infra/domains/' + domain_id + '/groups/' + group_id

        dfw_rule['destination_groups'] = [get_group_path(group_id, domain_id)
                                          for group_id in
                                          dfw_rule['destination_groups']]
        dfw_rule['source_groups'] = [get_group_path(group_id, domain_id) for
                                     group_id in dfw_rule['source_groups']]
        child_rules = [{'resource_type': 'ChildRule', 'Rule': dfw_rule}]
        security_policy.update({'children': child_rules})
        child_security_policies = [{
            'resource_type': 'ChildSecurityPolicy',
            'SecurityPolicy': security_policy
        }]
        domain.update({'children': child_security_policies})
        child_domains = [{'resource_type': 'ChildDomain',
                         'Domain': domain}]
        expected_body = {'resource_type': 'Infra',
                         'children': child_domains}
        self.assert_infra_patch_call(expected_body)

    @mock.patch('vmware_nsxlib.v3.policy.core_defs.NsxPolicyApi.get')
    def test_updating_security_policy_and_dfw_rules(self, mock_get_api):
        dfw_rule1 = {'id': 'rule_id1', 'action': 'ALLOW',
                     'display_name': 'rule1', 'description': None,
                     'direction': 'IN_OUT', 'ip_protocol': 'IPV4_IPV6',
                     'logged': False,
                     'destination_groups': ['destination_url'],
                     'source_groups': ['src_url'], 'resource_type': 'Rule',
                     'scope': None, 'sequence_number': None, 'tag': None,
                     'services': ['ANY'], "_create_time": 1}
        dfw_rule2 = {'id': 'rule_id2', 'action': 'DROP',
                     'display_name': 'rule2', 'description': None,
                     'direction': 'IN_OUT', 'ip_protocol': 'IPV4_IPV6',
                     'logged': False,
                     'destination_groups': ['destination_url'],
                     'source_groups': ['src_url'], 'resource_type': 'Rule',
                     'scope': None, 'sequence_number': None, 'tag': None,
                     'services': ['ANY'], "_create_time": 1}
        security_policy = {'id': 'security_policy_id1',
                           'display_name': 'security_policy',
                           'category': 'Application',
                           'resource_type': 'SecurityPolicy'}
        domain = {'resource_type': 'Domain', 'id': 'domain1'}
        domain_id = domain['id']
        map_id = security_policy['id']
        new_rule_name = 'new_rule1'
        new_direction = 'IN'
        dfw_rule_entries = [self.policy_lib.comm_map.build_entry(
            name=new_rule_name,
            domain_id=domain_id,
            map_id=map_id,
            entry_id=dfw_rule1['id'],
            source_groups=dfw_rule1['source_groups'],
            dest_groups=dfw_rule1['destination_groups'],
            direction=new_direction
        )]

        def get_group_path(group_id, domain_id):
            return '/infra/domains/' + domain_id + '/groups/' + group_id

        for dfw_rule in [dfw_rule1, dfw_rule2]:
            dfw_rule['destination_groups'] = [get_group_path(group_id,
                                                             domain_id)
                                              for group_id in
                                              dfw_rule['destination_groups']]
            dfw_rule['source_groups'] = [get_group_path(group_id, domain_id)
                                         for group_id in
                                         dfw_rule['source_groups']]

        security_policy_values = copy.deepcopy(security_policy)
        security_policy_values.update({'rules':
                                      copy.deepcopy([dfw_rule1, dfw_rule2])})
        mock_get_api.return_value = security_policy_values

        with trans.NsxPolicyTransaction():
            self.policy_lib.comm_map.update_with_entries(
                name=security_policy['display_name'],
                domain_id=domain_id,
                map_id=map_id,
                entries=dfw_rule_entries
            )

        dfw_rule1['display_name'] = new_rule_name
        dfw_rule1['direction'] = new_direction
        child_rules = [{'resource_type': 'ChildRule', 'Rule': dfw_rule1},
                       {'resource_type': 'ChildRule', 'Rule': dfw_rule2,
                        'marked_for_delete': True}]
        security_policy.update({'children': child_rules})
        child_security_policies = [{
            'resource_type': 'ChildSecurityPolicy',
            'SecurityPolicy': security_policy
        }]
        domain.update({'children': child_security_policies})
        child_domains = [{
            'resource_type': 'ChildDomain',
            'Domain': domain
        }]
        expected_body = {'resource_type': 'Infra',
                         'children': child_domains}
        self.assert_infra_patch_call(expected_body)

    @mock.patch('vmware_nsxlib.v3.policy.core_defs.NsxPolicyApi.get')
    def test_updating_security_policy_with_no_entries_set(self, mock_get_api):
        dfw_rule1 = {'id': 'rule_id1', 'action': 'ALLOW',
                     'display_name': 'rule1', 'description': None,
                     'direction': 'IN_OUT', 'ip_protocol': 'IPV4_IPV6',
                     'logged': False,
                     'destination_groups': ['destination_url'],
                     'source_groups': ['src_url'], 'resource_type': 'Rule',
                     'scope': None, 'sequence_number': None, 'tag': None,
                     'services': ['ANY'], "_create_time": 1}
        security_policy = {'id': 'security_policy_id1',
                           'display_name': 'security_policy',
                           'category': 'Application',
                           'resource_type': 'SecurityPolicy'}
        domain = {'resource_type': 'Domain', 'id': 'domain1'}
        domain_id = domain['id']
        map_id = security_policy['id']

        def get_group_path(group_id, domain_id):
            return '/infra/domains/' + domain_id + '/groups/' + group_id

        for dfw_rule in [dfw_rule1]:
            dfw_rule['destination_groups'] = [get_group_path(group_id,
                                                             domain_id)
                                              for group_id in
                                              dfw_rule['destination_groups']]
            dfw_rule['source_groups'] = [get_group_path(group_id, domain_id)
                                         for group_id in
                                         dfw_rule['source_groups']]

        security_policy.update({'rules': [dfw_rule1]})
        mock_get_api.return_value = security_policy

        with trans.NsxPolicyTransaction():
            self.policy_lib.comm_map.update_with_entries(
                name=security_policy['display_name'],
                domain_id=domain_id,
                map_id=map_id
            )

        child_security_policies = [{
            'resource_type': 'ChildSecurityPolicy',
            'SecurityPolicy': security_policy
        }]
        domain.update({'children': child_security_policies})
        child_domains = [{
            'resource_type': 'ChildDomain',
            'Domain': domain
        }]
        expected_body = {'resource_type': 'Infra',
                         'children': child_domains}
        self.assert_infra_patch_call(expected_body)



# Copyright (c) 2015 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from vmware_nsxlib.tests.unit.v3 import nsxlib_testcase
from vmware_nsxlib.v3 import exceptions
from vmware_nsxlib.v3 import nsx_constants
from vmware_nsxlib.v3 import utils


class TestNsxV3Utils(nsxlib_testcase.NsxClientTestCase):

    def test_build_v3_tags_payload(self):
        result = self.nsxlib.build_v3_tags_payload(
            {'id': 'fake_id',
             'project_id': 'fake_proj_id'},
            resource_type='os-net-id',
            project_name='fake_proj_name')
        expected = [{'scope': 'os-net-id', 'tag': 'fake_id'},
                    {'scope': 'os-project-id', 'tag': 'fake_proj_id'},
                    {'scope': 'os-project-name', 'tag': 'fake_proj_name'},
                    {'scope': 'os-api-version',
                     'tag': nsxlib_testcase.PLUGIN_VER}]
        self.assertEqual(expected, result)

    def test_build_v3_tags_payload_internal(self):
        result = self.nsxlib.build_v3_tags_payload(
            {'id': 'fake_id',
             'project_id': 'fake_proj_id'},
            resource_type='os-net-id',
            project_name=None)
        expected = [{'scope': 'os-net-id', 'tag': 'fake_id'},
                    {'scope': 'os-project-id', 'tag': 'fake_proj_id'},
                    {'scope': 'os-project-name',
                     'tag': nsxlib_testcase.PLUGIN_TAG},
                    {'scope': 'os-api-version',
                     'tag': nsxlib_testcase.PLUGIN_VER}]
        self.assertEqual(expected, result)

    def test_build_v3_tags_payload_invalid_length(self):
        self.assertRaises(exceptions.NsxLibInvalidInput,
                          self.nsxlib.build_v3_tags_payload,
                          {'id': 'fake_id',
                           'project_id': 'fake_proj_id'},
                          resource_type='os-longer-maldini-rocks-id',
                          project_name='fake')

    def test_build_v3_api_version_tag(self):
        result = self.nsxlib.build_v3_api_version_tag()
        expected = [{'scope': nsxlib_testcase.PLUGIN_SCOPE,
                     'tag': nsxlib_testcase.PLUGIN_TAG},
                    {'scope': 'os-api-version',
                     'tag': nsxlib_testcase.PLUGIN_VER}]
        self.assertEqual(expected, result)

    def test_build_v3_api_version_project_tag(self):
        proj = 'project_x'
        result = self.nsxlib.build_v3_api_version_project_tag(proj)
        expected = [{'scope': nsxlib_testcase.PLUGIN_SCOPE,
                     'tag': nsxlib_testcase.PLUGIN_TAG},
                    {'scope': 'os-api-version',
                     'tag': nsxlib_testcase.PLUGIN_VER},
                    {'scope': 'os-project-name',
                     'tag': proj}]
        self.assertEqual(expected, result)

    def test_build_v3_api_version_project_id_tag(self):
        proj = 'project_x'
        proj_id = 'project_id'
        result = self.nsxlib.build_v3_api_version_project_tag(
            proj, project_id=proj_id)
        expected = [{'scope': nsxlib_testcase.PLUGIN_SCOPE,
                     'tag': nsxlib_testcase.PLUGIN_TAG},
                    {'scope': 'os-api-version',
                     'tag': nsxlib_testcase.PLUGIN_VER},
                    {'scope': 'os-project-name',
                     'tag': proj},
                    {'scope': 'os-project-id',
                     'tag': proj_id}]
        self.assertEqual(expected, result)

    def test_is_internal_resource(self):
        project_tag = self.nsxlib.build_v3_tags_payload(
            {'id': 'fake_id',
             'project_id': 'fake_proj_id'},
            resource_type='os-net-id',
            project_name=None)
        internal_tag = self.nsxlib.build_v3_api_version_tag()

        expect_false = self.nsxlib.is_internal_resource({'tags': project_tag})
        self.assertFalse(expect_false)

        expect_true = self.nsxlib.is_internal_resource({'tags': internal_tag})
        self.assertTrue(expect_true)

    def test_get_name_and_uuid(self):
        uuid = 'afc40f8a-4967-477e-a17a-9d560d1786c7'
        suffix = '_afc40...786c7'
        expected = 'maldini%s' % suffix
        short_name = utils.get_name_and_uuid('maldini', uuid)
        self.assertEqual(expected, short_name)

        name = 'X' * 255
        expected = '%s%s' % ('X' * (80 - len(suffix)), suffix)
        short_name = utils.get_name_and_uuid(name, uuid)
        self.assertEqual(expected, short_name)

    def test_build_v3_tags_max_length_payload(self):
        result = self.nsxlib.build_v3_tags_payload(
            {'id': 'X' * 255,
             'project_id': 'X' * 255},
            resource_type='os-net-id',
            project_name='X' * 255)
        expected = [{'scope': 'os-net-id', 'tag': 'X' * 40},
                    {'scope': 'os-project-id', 'tag': 'X' * 40},
                    {'scope': 'os-project-name', 'tag': 'X' * 40},
                    {'scope': 'os-api-version',
                     'tag': nsxlib_testcase.PLUGIN_VER}]
        self.assertEqual(expected, result)

    def test_add_v3_tag(self):
        result = utils.add_v3_tag([], 'fake-scope', 'fake-tag')
        expected = [{'scope': 'fake-scope', 'tag': 'fake-tag'}]
        self.assertEqual(expected, result)

    def test_add_v3_tag_max_length_payload(self):
        result = utils.add_v3_tag([], 'fake-scope', 'X' * 255)
        expected = [{'scope': 'fake-scope', 'tag': 'X' * 40}]
        self.assertEqual(expected, result)

    def test_add_v3_tag_invalid_scope_length(self):
        self.assertRaises(exceptions.NsxLibInvalidInput,
                          utils.add_v3_tag,
                          [],
                          'fake-scope-name-is-far-too-long',
                          'fake-tag')

    def test_update_v3_tags_addition(self):
        tags = [{'scope': 'os-net-id', 'tag': 'X' * 40},
                {'scope': 'os-project-id', 'tag': 'Y' * 40},
                {'scope': 'os-project-name', 'tag': 'Z' * 40},
                {'scope': 'os-api-version',
                 'tag': nsxlib_testcase.PLUGIN_VER}]
        resources = [{'scope': 'os-instance-uuid',
                      'tag': 'A' * 40}]
        tags = utils.update_v3_tags(tags, resources)
        expected = [{'scope': 'os-net-id', 'tag': 'X' * 40},
                    {'scope': 'os-project-id', 'tag': 'Y' * 40},
                    {'scope': 'os-project-name', 'tag': 'Z' * 40},
                    {'scope': 'os-api-version',
                     'tag': nsxlib_testcase.PLUGIN_VER},
                    {'scope': 'os-instance-uuid',
                     'tag': 'A' * 40}]
        self.assertEqual(sorted(expected, key=lambda x: x.get('tag')),
                         sorted(tags, key=lambda x: x.get('tag')))

    def test_update_v3_tags_removal(self):
        tags = [{'scope': 'os-net-id', 'tag': 'X' * 40},
                {'scope': 'os-project-id', 'tag': 'Y' * 40},
                {'scope': 'os-project-name', 'tag': 'Z' * 40},
                {'scope': 'os-api-version',
                 'tag': nsxlib_testcase.PLUGIN_VER}]
        resources = [{'scope': 'os-net-id',
                      'tag': ''}]
        tags = utils.update_v3_tags(tags, resources)
        expected = [{'scope': 'os-project-id', 'tag': 'Y' * 40},
                    {'scope': 'os-project-name', 'tag': 'Z' * 40},
                    {'scope': 'os-api-version',
                     'tag': nsxlib_testcase.PLUGIN_VER}]
        self.assertEqual(sorted(expected, key=lambda x: x.get('tag')),
                         sorted(tags, key=lambda x: x.get('tag')))

    def test_update_v3_tags_update(self):
        tags = [{'scope': 'os-net-id', 'tag': 'X' * 40},
                {'scope': 'os-project-id', 'tag': 'Y' * 40},
                {'scope': 'os-project-name', 'tag': 'Z' * 40},
                {'scope': 'os-api-version',
                 'tag': nsxlib_testcase.PLUGIN_VER}]
        resources = [{'scope': 'os-project-id',
                      'tag': 'A' * 40}]
        tags = utils.update_v3_tags(tags, resources)
        expected = [{'scope': 'os-net-id', 'tag': 'X' * 40},
                    {'scope': 'os-project-id', 'tag': 'A' * 40},
                    {'scope': 'os-project-name', 'tag': 'Z' * 40},
                    {'scope': 'os-api-version',
                     'tag': nsxlib_testcase.PLUGIN_VER}]
        self.assertEqual(sorted(expected, key=lambda x: x.get('tag')),
                         sorted(tags, key=lambda x: x.get('tag')))

    def test_update_v3_tags_repetitive_scopes(self):
        tags = [{'scope': 'os-net-id', 'tag': 'X' * 40},
                {'scope': 'os-project-id', 'tag': 'Y' * 40},
                {'scope': 'os-project-name', 'tag': 'Z' * 40},
                {'scope': 'os-security-group', 'tag': 'SG1'},
                {'scope': 'os-security-group', 'tag': 'SG2'}]
        tags_update = [{'scope': 'os-security-group', 'tag': 'SG3'},
                       {'scope': 'os-security-group', 'tag': 'SG4'}]
        tags = utils.update_v3_tags(tags, tags_update)
        expected = [{'scope': 'os-net-id', 'tag': 'X' * 40},
                    {'scope': 'os-project-id', 'tag': 'Y' * 40},
                    {'scope': 'os-project-name', 'tag': 'Z' * 40},
                    {'scope': 'os-security-group', 'tag': 'SG3'},
                    {'scope': 'os-security-group', 'tag': 'SG4'}]
        self.assertEqual(sorted(expected, key=lambda x: x.get('tag')),
                         sorted(tags, key=lambda x: x.get('tag')))

    def test_update_v3_tags_repetitive_scopes_remove(self):
        tags = [{'scope': 'os-net-id', 'tag': 'X' * 40},
                {'scope': 'os-project-id', 'tag': 'Y' * 40},
                {'scope': 'os-project-name', 'tag': 'Z' * 40},
                {'scope': 'os-security-group', 'tag': 'SG1'},
                {'scope': 'os-security-group', 'tag': 'SG2'}]
        tags_update = [{'scope': 'os-security-group', 'tag': None}]
        tags = utils.update_v3_tags(tags, tags_update)
        expected = [{'scope': 'os-net-id', 'tag': 'X' * 40},
                    {'scope': 'os-project-id', 'tag': 'Y' * 40},
                    {'scope': 'os-project-name', 'tag': 'Z' * 40}]
        self.assertEqual(sorted(expected, key=lambda x: x.get('tag')),
                         sorted(tags, key=lambda x: x.get('tag')))

    def test_build_extra_args_positive(self):
        extra_args = ['fall_count', 'interval', 'monitor_port',
                      'request_body', 'request_method', 'request_url',
                      'request_version', 'response_body',
                      'response_status_codes', 'rise_count', 'timeout']
        body = {'display_name': 'httpmonitor1',
                'description': 'my http monitor'}
        expected = {'display_name': 'httpmonitor1',
                    'description': 'my http monitor',
                    'interval': 5,
                    'rise_count': 3,
                    'fall_count': 3}
        resp = utils.build_extra_args(body, extra_args, interval=5,
                                      rise_count=3, fall_count=3)
        self.assertEqual(resp, expected)

    def test_build_extra_args_negative(self):
        extra_args = ['cookie_domain', 'cookie_fallback', 'cookie_garble',
                      'cookie_mode', 'cookie_name', 'cookie_path',
                      'cookie_time']
        body = {'display_name': 'persistenceprofile1',
                'description': 'my persistence profile',
                'resource_type': 'LoadBalancerCookiePersistenceProfile'}
        expected = {'display_name': 'persistenceprofile1',
                    'description': 'my persistence profile',
                    'resource_type': 'LoadBalancerCookiePersistenceProfile',
                    'cookie_mode': 'INSERT',
                    'cookie_name': 'ABC',
                    'cookie_fallback': True}
        resp = utils.build_extra_args(body, extra_args, cookie_mode='INSERT',
                                      cookie_name='ABC', cookie_fallback=True,
                                      bogus='bogus')
        self.assertEqual(resp, expected)

    def test_retry(self):
        max_retries = 5
        total_count = {'val': 0}

        @utils.retry_upon_exception(exceptions.NsxLibInvalidInput,
                                    max_attempts=max_retries)
        def func_to_fail(x):
            total_count['val'] = total_count['val'] + 1
            raise exceptions.NsxLibInvalidInput(error_message='foo')

        self.assertRaises(exceptions.NsxLibInvalidInput, func_to_fail, 99)
        self.assertEqual(max_retries, total_count['val'])

    def test_retry_random(self):
        max_retries = 5
        total_count = {'val': 0}

        @utils.retry_random_upon_exception(exceptions.NsxLibInvalidInput,
                                           max_attempts=max_retries)
        def func_to_fail(x):
            total_count['val'] = total_count['val'] + 1
            raise exceptions.NsxLibInvalidInput(error_message='foo')

        self.assertRaises(exceptions.NsxLibInvalidInput, func_to_fail, 99)
        self.assertEqual(max_retries, total_count['val'])

    def test_retry_random_tuple(self):
        max_retries = 5
        total_count = {'val': 0}

        @utils.retry_random_upon_exception(
            (exceptions.NsxLibInvalidInput, exceptions.APITransactionAborted),
            max_attempts=max_retries)
        def func_to_fail(x):
            total_count['val'] = total_count['val'] + 1
            raise exceptions.NsxLibInvalidInput(error_message='foo')

        self.assertRaises(exceptions.NsxLibInvalidInput, func_to_fail, 99)
        self.assertEqual(max_retries, total_count['val'])

    @mock.patch.object(utils, '_update_max_nsgroups_criteria_tags')
    @mock.patch.object(utils, '_update_max_tags')
    @mock.patch.object(utils, '_update_tag_length')
    @mock.patch.object(utils, '_update_resource_length')
    def test_update_limits(self, _update_resource_length,
                           _update_tag_length, _update_max_tags,
                           _update_msx_nsg_criteria):
        limits = utils.TagLimits(1, 2, 3)
        utils.update_tag_limits(limits)
        _update_resource_length.assert_called_with(1)
        _update_tag_length.assert_called_with(2)
        _update_max_tags.assert_called_with(3)
        _update_msx_nsg_criteria.assert_called_with(3)


class NsxFeaturesTestCase(nsxlib_testcase.NsxLibTestCase):

    def test_v2_features(self, current_version='2.0.0'):
        self.nsxlib.nsx_version = current_version
        self.assertTrue(self.nsxlib.feature_supported(
            nsx_constants.FEATURE_ROUTER_FIREWALL))
        self.assertTrue(self.nsxlib.feature_supported(
            nsx_constants.FEATURE_EXCLUDE_PORT_BY_TAG))

    def test_v2_features_plus(self):
        self.test_v2_features(current_version='2.0.1')

    def test_v2_features_minus(self):
        self.nsxlib.nsx_version = '1.9.9'
        self.assertFalse(self.nsxlib.feature_supported(
            nsx_constants.FEATURE_ROUTER_FIREWALL))
        self.assertFalse(self.nsxlib.feature_supported(
            nsx_constants.FEATURE_EXCLUDE_PORT_BY_TAG))
        self.assertTrue(self.nsxlib.feature_supported(
            nsx_constants.FEATURE_MAC_LEARNING))



# Copyright 2019 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six


# NOTE: Consider inheriting from an abstract TokenProvider class to share
#       interface with XSRF token
@six.add_metaclass(abc.ABCMeta)
class AbstractJWTProvider(object):
    """Interface for providers of JSON Web Tokens(JWT)

    Responsible to provide the token value and refresh it once expired,
    or on demand, for authorization of requests to NSX.
    """

    @abc.abstractmethod
    def get_token(self, refresh_token=False):
        """Request JWT value.

        :param refresh_token: Boolean value, indicating whether a new token
                              value is to be retrieved.
        :raises vmware_nsxlib.v3.exceptions.BadJSONWebTokenProviderRequest:
        """
        pass

    def get_header_value(self, token_value):
        return "Bearer %s" % token_value



# Copyright 2016 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import inspect
import re
import time

from oslo_log import log
import tenacity
from tenacity import _utils as tenacity_utils

from vmware_nsxlib._i18n import _
from vmware_nsxlib.v3 import constants
from vmware_nsxlib.v3 import exceptions as nsxlib_exceptions
from vmware_nsxlib.v3 import nsx_constants

LOG = log.getLogger(__name__)

TagLimits = collections.namedtuple('TagLimits',
                                   ['scope_length', 'tag_length', 'max_tags'])

# These defaults reflect latest tag & scope limits on the backend. As of 2.5,
# backend no longer returns tag limit via API.
MAX_RESOURCE_TYPE_LEN = 128
MAX_TAG_LEN = 256
MAX_TAGS = 15
MAX_NSGROUPS_CRITERIA_TAGS = 10

DEFAULT_MAX_ATTEMPTS = 10
DEFAULT_CACHE_AGE_SEC = 600
INJECT_HEADERS_CALLBACK = None
IS_ATTR_SET_CALLBACK = None


def set_is_attr_callback(callback):
    global IS_ATTR_SET_CALLBACK
    IS_ATTR_SET_CALLBACK = callback


def is_attr_set(attr):
    if IS_ATTR_SET_CALLBACK:
        return IS_ATTR_SET_CALLBACK(attr)
    return attr is not None


def set_inject_headers_callback(callback):
    global INJECT_HEADERS_CALLBACK
    INJECT_HEADERS_CALLBACK = callback


def censor_headers(headers):
    censored_headers = ['authorization']
    result = {}
    for name, value in headers.items():
        if name.lower() in censored_headers:
            result[name] = '--- CENSORED ---'
        else:
            result[name] = value

    return result


def _update_resource_length(length):
    global MAX_RESOURCE_TYPE_LEN
    MAX_RESOURCE_TYPE_LEN = length


def _update_tag_length(length):
    global MAX_TAG_LEN
    MAX_TAG_LEN = length


def _update_max_tags(max_tags):
    global MAX_TAGS
    MAX_TAGS = max_tags


def _update_max_nsgroups_criteria_tags(max_tags):
    global MAX_NSGROUPS_CRITERIA_TAGS
    MAX_NSGROUPS_CRITERIA_TAGS = max(10, max_tags - 5)


def update_tag_limits(limits):
    _update_resource_length(limits.scope_length)
    _update_tag_length(limits.tag_length)
    _update_max_tags(limits.max_tags)
    _update_max_nsgroups_criteria_tags(limits.max_tags)


def _validate_resource_type_length(resource_type):
    # Add in a validation to ensure that we catch this at build time
    if len(resource_type) > MAX_RESOURCE_TYPE_LEN:
        raise nsxlib_exceptions.NsxLibInvalidInput(
            error_message=(_('Resource type cannot exceed %(max_len)s '
                             'characters: %(resource_type)s') %
                           {'max_len': MAX_RESOURCE_TYPE_LEN,
                            'resource_type': resource_type}))


def add_v3_tag(tags, resource_type, tag):
    _validate_resource_type_length(resource_type)
    tags.append({'scope': resource_type, 'tag': tag[:MAX_TAG_LEN]})
    return tags


def update_v3_tags(current_tags, tags_update):
    current_scopes = set([tag['scope'] for tag in current_tags])
    updated_scopes = set([tag['scope'] for tag in tags_update])

    # All tags scopes which are either completely new or already defined on the
    # resource are left in place, unless the tag value is empty, in that case
    # it is ignored.
    tags = [{'scope': tag['scope'], 'tag': tag['tag']}
            for tag in (current_tags + tags_update)
            if tag['tag'] and
            tag['scope'] in (current_scopes ^ updated_scopes)]

    modified_scopes = current_scopes & updated_scopes
    for tag in tags_update:
        if tag['scope'] in modified_scopes:
            # If the tag value is empty or None, then remove the tag completely
            if tag['tag']:
                tag['tag'] = tag['tag'][:MAX_TAG_LEN]
                tags.append(tag)

    return tags


def _log_before_retry(retry_state):
    """Before call strategy that logs to some logger the attempt."""
    if retry_state.attempt_number > 1:
        LOG.warning("Retrying call to '%(func)s' for the %(num)s time",
                    {'func': tenacity_utils.get_callback_name(
                        retry_state.fn),
                     'num': tenacity_utils.to_ordinal(
                        retry_state.attempt_number)})


def _get_args_from_frame(frames, frame_num):
    if len(frames) > frame_num and frames[frame_num] and frames[frame_num][0]:
        argvalues = inspect.getargvalues(frames[frame_num][0])
        formated_args = inspect.formatargvalues(*argvalues)
        # remove the first 'self' arg from the log as it adds no information
        formated_args = re.sub(r'\(self=.*?, ', "(", formated_args)
        return formated_args


def _log_after_retry(retry_state):
    """After call strategy that logs to some logger the finished attempt."""
    # Using inspect to get arguments of the relevant call
    frames = inspect.trace()
    # Look at frame #2 first because of the internal functions _do_X
    formated_args = _get_args_from_frame(frames, 2)
    if not formated_args:
        formated_args = _get_args_from_frame(frames, 1)
    if not formated_args:
        formated_args = "Unknown"

    LOG.warning("Finished retry of %(func)s for the %(num)s time after "
                "%(time)0.3f(s) with args: %(args)s",
                {'func': tenacity_utils.get_callback_name(retry_state.fn),
                 'num': tenacity_utils.to_ordinal(retry_state.attempt_number),
                 'time': retry_state.seconds_since_start,
                 'args': formated_args})


def retry_upon_exception(exc, delay=0.5, max_delay=2,
                         max_attempts=DEFAULT_MAX_ATTEMPTS):
    return tenacity.retry(reraise=True,
                          retry=tenacity.retry_if_exception_type(exc),
                          wait=tenacity.wait_exponential(
                              multiplier=delay, max=max_delay),
                          stop=tenacity.stop_after_attempt(max_attempts),
                          before=_log_before_retry, after=_log_after_retry)


def retry_random_upon_exception(exc, delay=0.5, max_delay=5,
                                max_attempts=DEFAULT_MAX_ATTEMPTS):
    return tenacity.retry(reraise=True,
                          retry=tenacity.retry_if_exception_type(exc),
                          wait=tenacity.wait_random_exponential(
                              multiplier=delay, max=max_delay),
                          stop=tenacity.stop_after_attempt(max_attempts),
                          before=_log_before_retry, after=_log_after_retry)


def retry_upon_none_result(max_attempts, delay=0.5, max_delay=10,
                           random=False):
    if random:
        wait_func = tenacity.wait_random_exponential(
            multiplier=delay, max=max_delay)
    else:
        wait_func = tenacity.wait_exponential(
            multiplier=delay, max=max_delay)
    return tenacity.retry(reraise=True,
                          retry=tenacity.retry_if_result(lambda x: x is None),
                          wait=wait_func,
                          stop=tenacity.stop_after_attempt(max_attempts),
                          before=_log_before_retry, after=_log_after_retry)


def list_match(list1, list2):
    # Check if list1 and list2 have identical elements, but relaxed on
    # dict elements where list1's dict element can be a subset of list2's
    # corresponding element.
    if (not isinstance(list1, list) or not isinstance(list2, list) or
        len(list1) != len(list2)):
        return False
    list1 = sorted(list1)
    list2 = sorted(list2)
    for (v1, v2) in zip(list1, list2):
        if isinstance(v1, dict):
            if not dict_match(v1, v2):
                return False
        elif isinstance(v1, list):
            if not list_match(v1, v2):
                return False
        elif v1 != v2:
            return False
    return True


def dict_match(dict1, dict2):
    # Check if dict1 is a subset of dict2.
    if not isinstance(dict1, dict) or not isinstance(dict2, dict):
        return False
    for k1, v1 in dict1.items():
        if k1 not in dict2:
            return False
        v2 = dict2[k1]
        if isinstance(v1, dict):
            if not dict_match(v1, v2):
                return False
        elif isinstance(v1, list):
            if not list_match(v1, v2):
                return False
        elif v1 != v2:
            return False
    return True


def get_name_and_uuid(name, uuid, tag=None, maxlen=80):
    short_uuid = '_' + uuid[:5] + '...' + uuid[-5:]
    maxlen = maxlen - len(short_uuid)
    if tag:
        maxlen = maxlen - len(tag) - 1
        return name[:maxlen] + '_' + tag + short_uuid
    else:
        return name[:maxlen] + short_uuid


def build_extra_args(body, extra_args, **kwargs):
    for arg in extra_args:
        if arg in kwargs:
            body[arg] = kwargs[arg]
    return body


def escape_tag_data(data):
    # ElasticSearch query_string requires slashes and dashes to
    # be escaped. We assume no other reserved characters will be
    # used in tag scopes or values
    return data.replace('/', '\\/').replace('-', '\\-').replace(':', '\\:')


def escape_display_name(display_name):
    # Illegal characters for the display names are  ;|=,~@
    rx = re.compile('([;|=,~@])')
    return rx.sub('.', display_name)


class NsxLibCache(object):
    def __init__(self, timeout):
        self.timeout = timeout
        self._cache = {}
        super(NsxLibCache, self).__init__()

    def expired(self, entry):
        return (time.time() - entry['time']) > self.timeout

    def get(self, key):
        if key in self._cache:
            # check that the value is still valid
            if self.expired(self._cache[key]):
                # this entry has expired
                self.remove(key)
            else:
                return self._cache[key]['value']

    def update(self, key, value):
        self._cache[key] = {'time': time.time(),
                            'value': value}

    def remove(self, key):
        if key in self._cache:
            del self._cache[key]


class NsxLibApiBase(object):
    """Base class for nsxlib api """
    def __init__(self, client, nsxlib_config=None, nsxlib=None):
        self.client = client
        self.nsxlib_config = nsxlib_config
        self.nsxlib = nsxlib
        super(NsxLibApiBase, self).__init__()
        self.cache = NsxLibCache(self.cache_timeout)

    @abc.abstractproperty
    def uri_segment(self):
        pass

    @abc.abstractproperty
    def resource_type(self):
        pass

    @property
    def use_cache_for_get(self):
        """By default no caching is used"""
        return False

    @property
    def cache_timeout(self):
        """the default cache aging time in seconds"""
        return DEFAULT_CACHE_AGE_SEC

    def get_path(self, resource=None):
        if resource:
            return '%s/%s' % (self.uri_segment, resource)
        return self.uri_segment

    def list(self):
        return self.client.list(self.uri_segment)

    def get(self, uuid, silent=False):
        if self.use_cache_for_get:
            # try to get it from the cache
            result = self.cache.get(uuid)
            if result:
                if not silent:
                    LOG.debug("Getting %s from cache.", self.get_path(uuid))
                return result
        # call the client
        result = self.client.get(self.get_path(uuid), silent=silent)
        if result and self.use_cache_for_get:
            # add the result to the cache
            self.cache.update(uuid, result)
        return result

    def read(self, uuid, silent=False):
        """The same as get"""
        return self.get(uuid, silent=silent)

    def delete(self, uuid):
        if self.use_cache_for_get:
            self.cache.remove(uuid)
        return self.client.delete(self.get_path(uuid))

    def find_by_display_name(self, display_name):
        found = []
        for resource in self.list()['results']:
            if resource['display_name'] == display_name:
                found.append(resource)
        return found

    def _update_with_retry(self, uuid, payload):
        if self.use_cache_for_get:
            self.cache.remove(uuid)
        return self._update_resource(self.get_path(uuid), payload, retry=True)

    def _internal_update_resource(self, resource, payload, headers=None,
                                  create_action=False,
                                  get_params=None,
                                  action_params=None,
                                  update_payload_cbk=None):
        get_path = action_path = resource
        if get_params:
            get_path = get_path + get_params
        if action_params:
            action_path = action_path + action_params
        revised_payload = self.client.get(get_path)
        # custom resource callback for updating the payload
        if update_payload_cbk:
            update_payload_cbk(revised_payload, payload)
        # special treatment for tags (merge old and new)
        if 'tags_update' in payload.keys():
            revised_payload['tags'] = update_v3_tags(
                revised_payload.get('tags', []),
                payload['tags_update'])
            del payload['tags_update']
        # update all the rest of the parameters
        for key_name in payload.keys():
            # handle 2 levels of dictionary:
            if isinstance(payload[key_name], dict):
                if key_name not in revised_payload:
                    revised_payload[key_name] = payload[key_name]
                else:
                    # copy each key
                    revised_payload[key_name].update(payload[key_name])
            else:
                revised_payload[key_name] = payload[key_name]
        if create_action:
            return self.client.create(action_path, revised_payload,
                                      headers=headers)
        else:
            return self.client.update(action_path, revised_payload,
                                      headers=headers)

    def _update_resource(self, resource, payload, headers=None,
                         create_action=False, get_params=None,
                         action_params=None, update_payload_cbk=None,
                         retry=False):
        if retry:
            # If revision_id of the payload that we send is older than what
            # NSX has, we will get a 412: Precondition Failed.
            # In that case we need to re-fetch, patch the response and send
            # it again with the new revision_id
            @retry_upon_exception(
                nsxlib_exceptions.StaleRevision,
                max_attempts=self.client.max_attempts)
            def do_update():
                return self._internal_update_resource(
                    resource, payload,
                    headers=headers,
                    create_action=create_action,
                    get_params=get_params,
                    action_params=action_params,
                    update_payload_cbk=update_payload_cbk)

            return do_update()
        else:
            return self._internal_update_resource(
                resource, payload,
                headers=headers,
                create_action=create_action,
                get_params=get_params,
                action_params=action_params,
                update_payload_cbk=update_payload_cbk)

    def _delete_with_retry(self, resource):
        # Using internal method so we can access max_attempts in the decorator
        @retry_upon_exception(
            nsxlib_exceptions.StaleRevision,
            max_attempts=self.client.max_attempts)
        def _do_delete():
            self.client.delete(self.get_path(resource))

        _do_delete()

    def _create_with_retry(self, resource, body=None, headers=None):
        # Using internal method so we can access max_attempts in the decorator
        @retry_upon_exception(
            nsxlib_exceptions.StaleRevision,
            max_attempts=self.client.max_attempts)
        def _do_create():
            return self.client.create(resource, body, headers=headers)

        return _do_create()

    def _get_resource_by_name_or_id(self, name_or_id, resource):
        all_results = self.client.list(resource)['results']
        matched_results = []
        for rs in all_results:
            if rs.get('id') == name_or_id:
                # Matched by id - must be unique
                return name_or_id

            if rs.get('display_name') == name_or_id:
                # Matched by name - add to the list to verify it is unique
                matched_results.append(rs)

        if len(matched_results) == 0:
            err_msg = (_("Could not find %(resource)s %(name)s") %
                       {'name': name_or_id, 'resource': resource})
            raise nsxlib_exceptions.ManagerError(details=err_msg)
        elif len(matched_results) > 1:
            err_msg = (_("Found multiple %(resource)s named %(name)s") %
                       {'name': name_or_id, 'resource': resource})
            raise nsxlib_exceptions.ManagerError(details=err_msg)

        return matched_results[0].get('id')

    def get_id_by_name_or_id(self, name_or_id):
        """Get a resource by it's display name or uuid

        Return the resource data, or raise an exception if not found or
        not unique
        """

        return self._get_resource_by_name_or_id(name_or_id,
                                                self.get_path())

    def build_v3_api_version_tag(self):
        """Some resources are created on the manager

        that do not have a corresponding plugin resource.

        """
        return [{'scope': self.nsxlib_config.plugin_scope,
                 'tag': self.nsxlib_config.plugin_tag},
                {'scope': "os-api-version",
                 'tag': self.nsxlib_config.plugin_ver}]

    def build_v3_api_version_project_tag(self, project_name, project_id=None):
        if not project_name:
            project_name = self.nsxlib_config.plugin_tag
        tags = [{'scope': self.nsxlib_config.plugin_scope,
                 'tag': self.nsxlib_config.plugin_tag},
                {'scope': "os-api-version",
                 'tag': self.nsxlib_config.plugin_ver},
                {'scope': 'os-project-name',
                 'tag': project_name[:MAX_TAG_LEN]}]
        if project_id:
            tags.append({'scope': 'os-project-id',
                         'tag': project_id[:MAX_TAG_LEN]})
        return tags

    def is_internal_resource(self, nsx_resource):
        """Indicates whether the passed nsx-resource is internal

        owned by the plugin for internal use.

        """
        for tag in nsx_resource.get('tags', []):
            if tag['scope'] == self.nsxlib_config.plugin_scope:
                return tag['tag'] == self.nsxlib_config.plugin_tag
        return False

    def build_v3_tags_payload(self, resource, resource_type, project_name):
        """Construct the tags payload that will be pushed to NSX-v3

        Add <resource_type>:<resource-id>, os-project-id:<project-id>,
        os-project-name:<project_name> os-api-version:<plugin-api-version>

        """
        _validate_resource_type_length(resource_type)
        # There may be cases when the plugin creates the port, for example DHCP
        if not project_name:
            project_name = self.nsxlib_config.plugin_tag
        project_id = (resource.get('project_id', '') or
                      resource.get('tenant_id', ''))
        # If project_id is present in resource and set to None, explicitly set
        # the project_id in tags as ''.
        if project_id is None:
            project_id = ''
        return [{'scope': resource_type,
                 'tag': resource.get('id', '')[:MAX_TAG_LEN]},
                {'scope': 'os-project-id',
                 'tag': project_id[:MAX_TAG_LEN]},
                {'scope': 'os-project-name',
                 'tag': project_name[:MAX_TAG_LEN]},
                {'scope': 'os-api-version',
                 'tag': self.nsxlib_config.plugin_ver}]


# Some utilities for services translations & validations
# both for the nsx manager & policy manager
def validate_icmp_params(icmp_type, icmp_code, icmp_version=4, strict=False):
    if icmp_version != 4:
        # ICMPv6 is currently not supported
        return
    if icmp_type:
        if (strict and icmp_type not in
                constants.IPV4_ICMP_STRICT_TYPES):
            raise nsxlib_exceptions.InvalidInput(
                operation='create_rule',
                arg_val=icmp_type,
                arg_name='icmp_type')
        if icmp_type not in constants.IPV4_ICMP_TYPES:
            raise nsxlib_exceptions.InvalidInput(
                operation='create_rule',
                arg_val=icmp_type,
                arg_name='icmp_type')
        if (icmp_code and strict and icmp_code not in
                constants.IPV4_ICMP_STRICT_TYPES[icmp_type]):
            raise nsxlib_exceptions.InvalidInput(
                operation='create_rule',
                arg_val=icmp_code,
                arg_name='icmp_code for this icmp_type')
        if (icmp_code and icmp_code not in
                constants.IPV4_ICMP_TYPES[icmp_type]):
            raise nsxlib_exceptions.InvalidInput(
                operation='create_rule',
                arg_val=icmp_code,
                arg_name='icmp_code for this icmp_type')


def get_l4_protocol_name(protocol_number):
    if protocol_number is None:
        return
    protocol_number = constants.IP_PROTOCOL_MAP.get(protocol_number,
                                                    protocol_number)
    try:
        protocol_number = int(protocol_number)
    except ValueError:
        raise nsxlib_exceptions.InvalidInput(
            operation='create_rule',
            arg_val=protocol_number,
            arg_name='protocol')
    if protocol_number == 6:
        return nsx_constants.TCP
    elif protocol_number == 17:
        return nsx_constants.UDP
    elif protocol_number == 1:
        return nsx_constants.ICMPV4
    else:
        return protocol_number


def get_dhcp_opt_code(name):
    _supported_options = {
        'subnet-mask': 1,
        'time-offset': 2,
        'router': 3,
        'dns-name': 6,
        'host-name': 12,
        'boot-file-size': 13,
        'domain-name': 15,
        'ip-forwarding': 19,
        'interface-mtu': 26,
        'broadcast-address': 28,
        'arp-cache-timeout': 35,
        'nis-domain': 40,
        'nis-servers': 41,
        'ntp-servers': 42,
        'netbios-name-servers': 44,
        'netbios-dd-server': 45,
        'netbios-node-type': 46,
        'netbios-scope': 47,
        'dhcp-renewal-time': 58,
        'dhcp-rebinding-time': 59,
        'class-id': 60,
        'dhcp-client-identifier': 61,
        'nisplus-domain': 64,
        'nisplus-servers': 65,
        'tftp-server': 66,
        'tftp-server-name': 66,
        'bootfile-name': 67,
        'system-architecture': 93,
        'interface-id': 94,
        'machine-id': 97,
        'name-search': 117,
        'subnet-selection': 118,
        'domain-search': 119,
        'classless-static-route': 121,
        'tftp-server-address': 150,
        'server-ip-address': 150,
        'etherboot': 175,
        'config-file': 209,
        'path-prefix': 210,
        'reboot-time': 211,
    }
    return _supported_options.get(name)



# Copyright 2016 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from time import time

from OpenSSL import crypto
from oslo_log import log
from oslo_utils import uuidutils

from vmware_nsxlib._i18n import _
from vmware_nsxlib.v3 import exceptions as nsxlib_exceptions

LOG = log.getLogger(__name__)

CERT_SUBJECT_COUNTRY = 'country'
CERT_SUBJECT_STATE = 'state'
CERT_SUBJECT_ORG = 'organization'
CERT_SUBJECT_UNIT = 'unit'
CERT_SUBJECT_HOST = 'hostname'


def validate_cert_params(key_size, valid_for_days,
                         signature_alg, subject):
    """Validate parameters for certificate"""

    expected_key_sizes = (2048, 4096)
    if key_size not in expected_key_sizes:
        raise nsxlib_exceptions.NsxLibInvalidInput(
            error_message=_('Invalid key size %(value)d'
                            '(must be one of %(list)s)') %
            {'value': key_size,
             'list': expected_key_sizes})

    expected_signature_algs = ('sha256')
    if signature_alg not in expected_signature_algs:
        raise nsxlib_exceptions.NsxLibInvalidInput(
            error_message=_('Invalid signature algorithm %(value)s'
                            '(must be one of %(list)s)') %
            {'value': signature_alg,
             'list': expected_signature_algs})

    if (CERT_SUBJECT_COUNTRY in subject and
       (len(subject[CERT_SUBJECT_COUNTRY]) != 2)):
        raise nsxlib_exceptions.NsxLibInvalidInput(
            error_message=_('Invalid country %s: '
                            'must be exactly 2 characters') %
            subject[CERT_SUBJECT_COUNTRY])

    # values defined in rfc5280
    max_len_constraints = {CERT_SUBJECT_STATE: 128,
                           CERT_SUBJECT_ORG: 64,
                           CERT_SUBJECT_UNIT: 64,
                           CERT_SUBJECT_HOST: 64}

    for field, max_len in max_len_constraints.items():
        if field in subject and (len(subject[field]) > max_len):
            raise nsxlib_exceptions.NsxLibInvalidInput(
                error_message=_('Invalid %(field)s [%(value)s]: '
                                'must not exceed %(max)d characters') %
                {'field': field,
                 'value': subject[field],
                 'max': max_len})


def generate_self_signed_cert_pair(key_size, valid_for_days,
                                   signature_alg, subject):
    """Generate self signed certificate and key pair"""

    validate_cert_params(key_size, valid_for_days,
                         signature_alg, subject)

    # generate key pair
    key = crypto.PKey()
    key.generate_key(crypto.TYPE_RSA, key_size)

    # generate certificate
    cert = crypto.X509()
    cert.get_subject().C = subject.get(CERT_SUBJECT_COUNTRY, 'US')
    cert.get_subject().ST = subject.get(CERT_SUBJECT_STATE, 'California')
    cert.get_subject().O = subject.get(CERT_SUBJECT_ORG, 'MyOrg')
    cert.get_subject().OU = subject.get(CERT_SUBJECT_UNIT, 'MyUnit')
    cert.get_subject().CN = subject.get(CERT_SUBJECT_HOST, 'myorg.com')
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(valid_for_days * 24 * 60 * 60)
    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(key)
    cert.set_serial_number(int(time()))
    cert.sign(key, signature_alg)

    return cert, key


class ClientCertificateManager(object):
    """Manage Client Certificate for backend authentication

    There should be single client certificate associated
    with certain principal identity. Certificate and PK storage
    is pluggable. Storage API (similar to neutron-lbaas barbican API):
        store_cert(purpose, certificate, private_key)
        get_cert(purpose)
        delete_cert(purpose)
    """

    def __init__(self, identity, nsx_trust_management, storage_driver):
        self._cert = None
        self._key = None
        self._storage_driver = storage_driver
        self._identity = identity

        self._nsx_trust_management = nsx_trust_management

    def __enter__(self):
        """Load cert from storage

        This is an optimization to avoid repeated storage access.
        Usage example:
            with cert_manager as c:
                if c.exists():
                    date = c.expires_on()
                    days = c.exires_in_days()
        """

        self._cert, self._key = self.get_cert_and_key()
        return self

    def __exit__(self, type, value, traceback):
        self._cert = None
        self._key = None

    def generate(self, subject, key_size=2048, valid_for_days=3650,
                 signature_alg='sha256', node_id=None):
        """Generate new certificate and register it in the system

        Generate certificate with RSA key based on arguments provided,
        register and associate it to principal identity on backend,
        and store it in storage. If certificate already exists, fail.
        """
        self._validate_empty()

        cert, key = generate_self_signed_cert_pair(key_size,
                                                   valid_for_days,
                                                   signature_alg,
                                                   subject)

        # register on backend
        self._register_cert(cert, node_id or uuidutils.generate_uuid())

        # save in storage
        cert_pem = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)
        key_pem = crypto.dump_privatekey(crypto.FILETYPE_PEM, key)
        self._storage_driver.store_cert(self._identity, cert_pem, key_pem)

        LOG.debug("Client certificate generated successfully")

    def delete(self):
        """Delete existing certificate from storage and backend"""
        cert_pem, key_pem = self.get_pem()
        if not cert_pem:
            return

        ok = True
        try:
            self._nsx_trust_management.delete_cert_and_identity(
                self._identity, cert_pem)
        except nsxlib_exceptions.ManagerError as e:
            LOG.error("Failed to clear certificate on backend: %s", e)
            ok = False

        try:
            self._storage_driver.delete_cert(self._identity)
        except Exception as e:
            LOG.error("Failed to clear certificate in storage: %s", e)
            ok = False

        self._cert = None
        self._key = None

        if ok:
            LOG.debug("Client certificate removed successfully")

    def exists(self):
        """Check if certificate was created for given identity"""
        if self._cert:
            return True

        cert_pem, key_pem = self._storage_driver.get_cert(self._identity)
        return cert_pem is not None

    def _get_cert_from_file(self, filename):
        with open(filename, 'r') as f:
            cert_pem = f.read()

        if not cert_pem:
            raise nsxlib_exceptions.CertificateError(
                msg=_("Failed to read certificate from %s") % filename)

        # validate correct crypto
        try:
            cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert_pem)
        except crypto.Error:
            raise nsxlib_exceptions.CertificateError(
                msg=_("Failed to import client certificate"))

        return cert

    def import_pem(self, filename, node_id=None):
        """Import and register existing certificate in PEM format"""

        # TODO(annak): support PK import as well
        self._validate_empty()

        cert = self._get_cert_from_file(filename)
        # register on backend
        self._register_cert(cert, node_id or uuidutils.generate_uuid())

        cert_pem = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)
        self._storage_driver.store_cert(self._identity, cert_pem, None)

        LOG.debug("Client certificate imported successfully")

    def delete_pem(self, filename):
        """Delete specified client certificate without storage verification"""
        # This file may contain private key
        # passing the pem through crypto will perform validation and
        # strip off the key
        cert = self._get_cert_from_file(filename)
        cert_pem = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)

        self._nsx_trust_management.delete_cert_and_identity(self._identity,
                                                            cert_pem)
        self._storage_driver.delete_cert(self._identity)

    def _load_from_storage(self):
        """Returns certificate and key pair in PEM format"""

        cert_pem, key_pem = self._storage_driver.get_cert(self._identity)
        if cert_pem is None:
            return None, None

        return (cert_pem, key_pem)

    def get_pem(self):
        return self._load_from_storage()

    def export_pem(self, filename):
        """Exports certificate and key pair to file"""
        self._validate_exists()

        cert_pem, key_pem = self._load_from_storage()

        with open(filename, 'w') as f:
            f.write(cert_pem)
            f.write(key_pem)

    def expires_on(self):
        """Returns certificate expiration timestamp"""
        self._validate_exists()

        cert, key = self.get_cert_and_key()
        converted = datetime.datetime.strptime(
            cert.get_notAfter().decode(),
            "%Y%m%d%H%M%SZ")
        return converted

    def expires_in_days(self):
        """Returns in how many days the certificate expires"""
        delta = self.expires_on() - datetime.datetime.utcnow()
        return delta.days

    def get_subject(self):
        self._validate_exists()

        cert, key = self.get_cert_and_key()
        return {CERT_SUBJECT_COUNTRY: cert.get_subject().C,
                CERT_SUBJECT_STATE: cert.get_subject().ST,
                CERT_SUBJECT_ORG: cert.get_subject().O,
                CERT_SUBJECT_UNIT: cert.get_subject().OU,
                CERT_SUBJECT_HOST: cert.get_subject().CN}

    def get_signature_alg(self):
        self._validate_exists()

        cert, key = self.get_cert_and_key()
        return cert.get_signature_algorithm()

    def get_key_size(self):
        self._validate_exists()

        cert, key = self.get_cert_and_key()
        return key.bits()

    def _validate_empty(self):
        if self.exists():
            raise nsxlib_exceptions.ObjectAlreadyExists(
                object_type='Client Certificate')

    def _validate_exists(self):
        if not self.exists():
            raise nsxlib_exceptions.ObjectNotGenerated(
                object_type='Client Certificate')

    def get_cert_and_key(self):
        """Load cert and key from storage"""
        if self._cert and self._key:
            return self._cert, self._key

        cert_pem, key_pem = self._load_from_storage()

        if cert_pem is None:
            return None, None

        try:
            cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert_pem)
            key = crypto.load_privatekey(crypto.FILETYPE_PEM, key_pem)
        except crypto.Error:
            raise nsxlib_exceptions.CertificateError(
                msg="Failed to load client certificate")

        return cert, key

    def _register_cert(self, cert, node_id):
        cert_pem = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)

        self._nsx_trust_management.create_cert_and_identity(self._identity,
                                                            cert_pem,
                                                            node_id)


class ClientCertProvider(object):
    """Basic implementation for client certificate provider

       Responsible for preparing, providing and disposing client certificate
       file. Basic implementation assumes the file exists in the file system
       and does not take responsibility of deleting this sensitive information
       after use.
       Inheriting objects should make use of __enter__ and __exit__ APIs to
       prepare and dispose the certificate file data.
    """
    def __init__(self, filename):
        self._filename = filename

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def filename(self):
        return self._filename



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging
from vmware_nsxlib.v3 import nsx_constants
from vmware_nsxlib.v3.policy import constants
from vmware_nsxlib.v3.policy.core_defs import ResourceDef

LOG = logging.getLogger(__name__)

TENANTS_PATH_PATTERN = "%s/"
LB_VIRTUAL_SERVERS_PATH_PATTERN = TENANTS_PATH_PATTERN + "lb-virtual-servers/"
LB_SERVICES_PATH_PATTERN = TENANTS_PATH_PATTERN + "lb-services/"
LB_POOL_PATH_PATTERN = TENANTS_PATH_PATTERN + "lb-pools/"
LB_APP_PROFILE_PATTERN = TENANTS_PATH_PATTERN + "lb-app-profiles/"
LB_MONITOR_PROFILE_PATTERN = TENANTS_PATH_PATTERN + "lb-monitor-profiles/"
LB_CLIENT_SSL_PROFILE_PATTERN = (TENANTS_PATH_PATTERN +
                                 "lb-client-ssl-profiles/")
LBSERVER_SSL_PROFILE_PATTERN = (TENANTS_PATH_PATTERN +
                                "lb-server-ssl-profiles/")
LB_PERSISTENCE_PROFILE_PATTERN = (TENANTS_PATH_PATTERN +
                                  "lb-persistence-profiles/")


class LBRuleDef(object):
    def __init__(self, actions, match_conditions=None, name=None,
                 match_strategy=None, phase=None):
        self.actions = actions
        self.name = name
        self.match_conditions = match_conditions
        self.match_strategy = match_strategy
        self.phase = phase

    def get_obj_dict(self):
        lb_rule = {
            'actions': self.actions
        }
        if self.match_conditions:
            lb_rule['match_conditions'] = self.match_conditions
        if self.name:
            lb_rule['display_name'] = self.name
        if self.match_strategy:
            lb_rule['match_strategy'] = self.match_strategy
        if self.phase:
            lb_rule['phase'] = self.phase
        return lb_rule


class LBPoolMemberDef(object):
    def __init__(self, ip_address, port=None, name=None,
                 weight=None, admin_state=None, backup_member=None):
        self.name = name
        self.ip_address = ip_address
        self.port = port
        self.weight = weight
        self.admin_state = admin_state
        self.backup_member = backup_member

    def get_obj_dict(self):
        body = {'ip_address': self.ip_address}
        if self.name:
            body['display_name'] = self.name
        if self.ip_address:
            body['port'] = self.port
        if self.weight:
            body['weight'] = self.weight
        if self.admin_state:
            body['admin_state'] = self.admin_state
        if self.backup_member:
            body['backup_member'] = self.backup_member
        return body


class LBServerSslProfileDef(ResourceDef):

    @property
    def path_pattern(self):
        return LBSERVER_SSL_PROFILE_PATTERN

    @property
    def path_ids(self):
        return ('tenant', 'server_ssl_profile_id')

    @staticmethod
    def resource_type():
        return "LBServerSslProfile"

    def get_obj_dict(self):
        body = super(LBServerSslProfileDef, self).get_obj_dict()
        self._set_attrs_if_specified(body, ['cipher_group_label', 'ciphers',
                                            'protocols',
                                            'session_cache_enabled'])
        return body


class LBClientSslProfileDef(ResourceDef):

    @property
    def path_pattern(self):
        return LB_CLIENT_SSL_PROFILE_PATTERN

    @property
    def path_ids(self):
        return ('tenant', 'client_ssl_profile_id')

    @staticmethod
    def resource_type():
        return "LBClientSslProfile"

    def get_obj_dict(self):
        body = super(LBClientSslProfileDef, self).get_obj_dict()
        self._set_attr_if_specified(body, 'protocols')
        return body


class LBPersistenceProfileBase(ResourceDef):

    @property
    def path_pattern(self):
        return LB_PERSISTENCE_PROFILE_PATTERN

    @property
    def path_ids(self):
        return ('tenant', 'persistence_profile_id')


class LBCookiePersistenceProfileDef(LBPersistenceProfileBase):

    @staticmethod
    def resource_type():
        return "LBCookiePersistenceProfile"

    def get_obj_dict(self):
        body = super(LBCookiePersistenceProfileDef, self).get_obj_dict()
        self._set_attrs_if_specified(
            body, ['cookie_garble', 'cookie_mode', 'cookie_name',
                   'cookie_path', 'cookie_time', 'persistence_shared'])
        return body


class LBSourceIpPersistenceProfileDef(LBPersistenceProfileBase):

    @staticmethod
    def resource_type():
        return "LBSourceIpPersistenceProfile"

    def get_obj_dict(self):
        body = super(LBSourceIpPersistenceProfileDef, self).get_obj_dict()
        self._set_attrs_if_specified(
            body, ['ha_persistence_mirroring_enabled', 'persistence_shared',
                   'purge', 'timeout'])
        return body


class LBAppProfileBaseDef(ResourceDef):

    @property
    def path_pattern(self):
        return LB_APP_PROFILE_PATTERN

    @property
    def path_ids(self):
        return ('tenant', 'lb_app_profile_id')

    def get_obj_dict(self):
        body = super(LBAppProfileBaseDef, self).get_obj_dict()
        self._set_attrs_if_specified(
            body, ['idle_timeout'])
        return body


class LBHttpProfileDef(LBAppProfileBaseDef):

    @staticmethod
    def resource_type():
        return "LBHttpProfile"

    def get_obj_dict(self):
        body = super(LBHttpProfileDef, self).get_obj_dict()
        self._set_attrs_if_specified(
            body, ['http_redirect_to', 'http_redirect_to_https', 'ntlm',
                   'request_body_size', 'request_header_size',
                   'response_header_size', 'response_timeout',
                   'x_forwarded_for'])
        return body


class LBFastTcpProfile(LBAppProfileBaseDef):

    @staticmethod
    def resource_type():
        return "LBFastTcpProfile"

    def get_obj_dict(self):
        body = super(LBFastTcpProfile, self).get_obj_dict()
        self._set_attrs_if_specified(
            body, ['close_timeout', 'ha_flow_mirroring_enabled'])
        return body


class LBFastUdpProfile(LBAppProfileBaseDef):

    @staticmethod
    def resource_type():
        return "LBFastUdpProfile"

    def get_obj_dict(self):
        body = super(LBFastUdpProfile, self).get_obj_dict()
        self._set_attrs_if_specified(
            body, ['flow_mirroring_enabled'])
        return body


class LBPoolDef(ResourceDef):

    @property
    def path_pattern(self):
        return LB_POOL_PATH_PATTERN

    @property
    def path_ids(self):
        return ('tenant', 'lb_pool_id')

    @staticmethod
    def resource_type():
        return 'LBPool'

    def get_obj_dict(self):
        body = super(LBPoolDef, self).get_obj_dict()
        self._set_attrs_if_specified(
            body, ['active_monitor_paths',
                   'algorithm', 'member_group', 'snat_translation'])
        members = self.get_attr('members')
        if members is None:
            members = []
        if self.has_attr('members'):
            members = members if isinstance(members, list) else [members]
            body['members'] = []
            for member in members:
                # the list contains old json members and newly added member
                if isinstance(member, LBPoolMemberDef):
                    member = member.get_obj_dict()
                body['members'].append(member)
        return body


class LBVirtualServerDef(ResourceDef):

    @property
    def path_pattern(self):
        return LB_VIRTUAL_SERVERS_PATH_PATTERN

    @property
    def path_ids(self):
        return ('tenant', 'virtual_server_id')

    @staticmethod
    def resource_type():
        return 'LBVirtualServer'

    def get_obj_dict(self):
        body = super(LBVirtualServerDef, self).get_obj_dict()
        self._set_attrs_if_specified(
            body, ['ip_address', 'ports', 'max_concurrent_connections'])
        client_ssl_binding = self.get_attr('client_ssl_profile_binding')
        if client_ssl_binding:
            self._set_attr_if_specified(
                body, 'client_ssl_profile_binding',
                value=client_ssl_binding)
        server_ssl_binding = self.get_attr('server_ssl_profile_binding')
        if server_ssl_binding:
            self._set_attr_if_specified(
                body, 'server_ssl_profile_binding',
                value=server_ssl_binding)
        waf_profile_binding = self.get_attr('waf_profile_binding')
        if waf_profile_binding:
            if isinstance(waf_profile_binding, WAFProfileBindingDef):
                waf_profile_binding = waf_profile_binding.get_obj_dict()
            self._set_attr_if_specified(
                body, 'waf_profile_binding',
                value=waf_profile_binding)
        rules = self.get_attr('rules')
        if self.has_attr('rules'):
            rules = rules if isinstance(rules, list) else [rules]
            body['rules'] = []
            for rule in rules:
                # the list contains old json rules and newly added ruledef rule
                if isinstance(rule, LBRuleDef):
                    rule = rule.get_obj_dict()
                body['rules'].append(rule)
        app_profile_id = self.get_attr('application_profile_id')
        if app_profile_id:
            app_profile_def = LBAppProfileBaseDef(
                lb_app_profile_id=app_profile_id, tenant=self.get_tenant())
            body['application_profile_path'] = (
                app_profile_def.get_resource_full_path())

        if self.has_attr('lb_persistence_profile_id'):
            path = ""
            lb_persistence_profile_id = self.get_attr(
                'lb_persistence_profile_id')
            if lb_persistence_profile_id:
                lb_persistence_profile_def = LBPersistenceProfileBase(
                    persistence_profile_id=lb_persistence_profile_id,
                    tenant=self.get_tenant())
                path = lb_persistence_profile_def.get_resource_full_path()
            body['lb_persistence_profile_path'] = path
        if self.has_attr('lb_service_id'):
            path = ""
            lb_service_id = self.get_attr('lb_service_id')
            if lb_service_id:
                lb_service_def = LBServiceDef(
                    lb_service_id=lb_service_id, tenant=self.get_tenant())
                path = lb_service_def.get_resource_full_path()
            body['lb_service_path'] = path
        if self.has_attr('pool_id'):
            path = ""
            lb_pool_id = self.get_attr('pool_id')
            if lb_pool_id:
                lb_pool_def = LBPoolDef(
                    lb_pool_id=lb_pool_id, tenant=self.get_tenant())
                path = lb_pool_def.get_resource_full_path()
            body['pool_path'] = path
        if self.has_attr('access_list_control'):
            lb_alc = self.get_attr('access_list_control')
            if isinstance(lb_alc, LBAccessListControlDef):
                self.attrs['access_list_control'] = lb_alc.get_obj_dict()
            self._set_attrs_if_supported(body, ['access_list_control'])
        return body

    @property
    def version_dependant_attr_map(self):
        return {'access_list_control': nsx_constants.NSX_VERSION_3_0_0}


class ClientSSLProfileBindingDef(object):
    def __init__(self, default_certificate_path, sni_certificate_paths=None,
                 ssl_profile_path=None, client_auth_ca_paths=None,
                 client_auth=None):
        self.default_certificate_path = default_certificate_path
        self.sni_certificate_paths = sni_certificate_paths
        self.ssl_profile_path = ssl_profile_path
        self.client_auth_ca_paths = client_auth_ca_paths
        self.client_auth = client_auth

    def get_obj_dict(self):
        body = {
            'default_certificate_path': self.default_certificate_path
        }
        if self.sni_certificate_paths:
            body['sni_certificate_paths'] = self.sni_certificate_paths
        if self.ssl_profile_path:
            body['ssl_profile_path'] = self.ssl_profile_path
        if self.client_auth_ca_paths:
            body['client_auth_ca_paths'] = self.client_auth_ca_paths
        if self.client_auth:
            body['client_auth'] = self.client_auth
        return body


class ServerSSLProfileBindingDef(object):
    def __init__(self, client_certificate_path=None,
                 certificate_chain_depth=None,
                 server_auth=None, server_auth_ca_paths=None,
                 server_auth_crl_paths=None, ssl_profile_path=None):
        self.client_certificate_path = client_certificate_path
        self.certificate_chain_depth = certificate_chain_depth
        self.server_auth = server_auth
        self.server_auth_ca_paths = server_auth_ca_paths
        self.server_auth_crl_paths = server_auth_crl_paths
        self.ssl_profile_path = ssl_profile_path

    def get_obj_dict(self):
        body = {}
        if self.client_certificate_path:
            body['client_certificate_path'] = self.client_certificate_path
        if self.ssl_profile_path:
            body['certificate_chain_depth'] = self.certificate_chain_depth
        if self.server_auth:
            body['server_auth'] = self.server_auth
        if self.ssl_profile_path:
            body['server_auth_ca_paths'] = self.server_auth_ca_paths
        if self.server_auth_crl_paths:
            body['server_auth_crl_paths'] = self.server_auth_crl_paths
        if self.ssl_profile_path:
            body['ssl_profile_path'] = self.ssl_profile_path
        return body


class WAFProfileBindingDef(object):
    def __init__(self, waf_profile_path,
                 operational_mode=constants.WAF_OPERATIONAL_MODE_PROTECTION,
                 debug_log_level=constants.WAF_LOG_LEVEL_NO_LOG):
        self.waf_profile_path = waf_profile_path
        self.operational_mode = operational_mode
        self.debug_log_level = debug_log_level

    def get_obj_dict(self):
        body = {
            'waf_profile_path': self.waf_profile_path,
            'operational_mode': self.operational_mode,
            'debug_log_level': self.debug_log_level
        }
        return body


class LBServiceDef(ResourceDef):

    @property
    def path_pattern(self):
        return LB_SERVICES_PATH_PATTERN

    @property
    def path_ids(self):
        return ('tenant', 'lb_service_id')

    @staticmethod
    def resource_type():
        return 'LBService'

    def get_obj_dict(self):
        body = super(LBServiceDef, self).get_obj_dict()
        self._set_attrs_if_specified(body, ['size', 'connectivity_path'])
        self._set_attrs_if_supported(body, ['relax_scale_validation'])
        return body

    @property
    def version_dependant_attr_map(self):
        return {'relax_scale_validation': nsx_constants.NSX_VERSION_3_0_0}


class LBServiceStatisticsDef(ResourceDef):

    @property
    def path_pattern(self):
        return LB_SERVICES_PATH_PATTERN + '%s/statistics/'

    @property
    def path_ids(self):
        return ('tenant', 'lb_service_id', '')


class LBServiceStatusDef(ResourceDef):

    @property
    def path_pattern(self):
        return LB_SERVICES_PATH_PATTERN + '%s/detailed-status/'

    @property
    def path_ids(self):
        return ('tenant', 'lb_service_id', '')


class LBServiceUsageDef(ResourceDef):

    def __init__(self, **kwargs):
        self.realtime = kwargs.pop('realtime')
        super(LBServiceUsageDef, self).__init__(**kwargs)

    @property
    def path_pattern(self):
        if self.realtime:
            return (LB_SERVICES_PATH_PATTERN +
                    '%s/service-usage?source=realtime')
        return LB_SERVICES_PATH_PATTERN + '%s/service-usage/'

    @property
    def path_ids(self):
        return ('tenant', 'lb_service_id', '')


class LBVirtualServerStatusDef(ResourceDef):

    @property
    def path_pattern(self):
        return (LB_SERVICES_PATH_PATTERN +
                '%s/lb-virtual-servers/%s/detailed-status/')

    @property
    def path_ids(self):
        return ('tenant', 'lb_service_id', 'lb_virtual_server_id', '')


class LBMonitorProfileBaseDef(ResourceDef):

    addl_attrs = ['interval', 'timeout', 'fall_count', 'rise_count']

    @property
    def path_pattern(self):
        return LB_MONITOR_PROFILE_PATTERN

    @property
    def path_ids(self):
        return ('tenant', 'lb_monitor_profile_id')

    def get_obj_dict(self):
        body = super(LBMonitorProfileBaseDef, self).get_obj_dict()
        self._set_attrs_if_specified(body, self.addl_attrs)
        return body


class LBHttpMonitorProfileDef(LBMonitorProfileBaseDef):

    addl_attrs = LBMonitorProfileBaseDef.addl_attrs + [
        'monitor_port', 'request_url', 'request_method', 'request_version',
        'request_headers', 'request_body', 'response_status_codes']

    @staticmethod
    def resource_type():
        return "LBHttpMonitorProfile"


class LBHttpsMonitorProfileDef(LBHttpMonitorProfileDef):

    @staticmethod
    def resource_type():
        return "LBHttpsMonitorProfile"


class LBUdpMonitorProfileDef(LBMonitorProfileBaseDef):

    addl_attrs = LBMonitorProfileBaseDef.addl_attrs + [
        'monitor_port', 'receive', 'send']

    @staticmethod
    def resource_type():
        return "LBUdpMonitorProfile"


class LBIcmpMonitorProfileDef(LBMonitorProfileBaseDef):

    @staticmethod
    def resource_type():
        return "LBIcmpMonitorProfile"


class LBTcpMonitorProfileDef(LBMonitorProfileBaseDef):

    addl_attrs = LBMonitorProfileBaseDef.addl_attrs + ['monitor_port']

    @staticmethod
    def resource_type():
        return "LBTcpMonitorProfile"


class LBAccessListControlDef(object):
    def __init__(self, action, group_path, enabled=None):
        self.action = action
        self.group_path = group_path
        self.enabled = enabled

    def get_obj_dict(self):
        access_list_control = {
            'action': self.action,
            'group_path': self.group_path
        }
        if self.enabled is not None:
            access_list_control['enabled'] = self.enabled
        return access_list_control



# Copyright 2016 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
import six

from oslo_log import log
from oslo_log import versionutils

from vmware_nsxlib.v3 import constants
from vmware_nsxlib.v3 import utils


LOG = log.getLogger(__name__)


class NsxLibNativeDhcp(utils.NsxLibApiBase):

    def build_static_routes(self, gateway_ip, cidr, host_routes):
        # The following code is based on _generate_opts_per_subnet() in
        # neutron/agent/linux/dhcp.py. It prepares DHCP options for a subnet.

        # Add route for directly connected network.
        static_routes = [{'network': cidr, 'next_hop': '0.0.0.0'}]
        # Copy routes from subnet host_routes attribute.
        if host_routes:
            for hr in host_routes:
                if hr['destination'] == constants.IPv4_ANY:
                    if not gateway_ip:
                        gateway_ip = hr['nexthop']
                else:
                    static_routes.append({'network': hr['destination'],
                                          'next_hop': hr['nexthop']})

        # If gateway_ip is defined, add default route via this gateway.
        if gateway_ip:
            static_routes.append({'network': constants.IPv4_ANY,
                                  'next_hop': gateway_ip})
        return static_routes, gateway_ip

    def build_server_name(self, net_name, net_id):
        return utils.get_name_and_uuid(net_name or 'dhcpserver', net_id)

    def build_server_domain_name(self, net_dns_domain, default_dns_domain):
        versionutils.report_deprecated_feature(
            LOG,
            'NsxLibQosNativeDhcp.build_server_domain_name is deprecated.')

        if net_dns_domain:
            if isinstance(net_dns_domain, six.string_types):
                domain_name = net_dns_domain
            else:
                domain_name = net_dns_domain['dns_domain']
        else:
            # use the default one, or the globally configured one
            if default_dns_domain is not None:
                domain_name = default_dns_domain
            else:
                domain_name = self.nsxlib_config.dns_domain
        return domain_name

    def build_server(self, name, ip_address, cidr, gateway_ip,
                     dns_domain=None, dns_nameservers=None,
                     host_routes=None,
                     dhcp_profile_id=None,
                     tags=None):

        # Prepare the configuration for a new logical DHCP server.
        server_ip = "%s/%u" % (ip_address,
                               netaddr.IPNetwork(cidr).prefixlen)

        if not dns_domain:
            dns_domain = self.nsxlib_config.dns_domain

        if not dns_nameservers:
            dns_nameservers = self.nsxlib_config.dns_nameservers

        if not utils.is_attr_set(gateway_ip):
            gateway_ip = None

        static_routes, gateway_ip = self.build_static_routes(
            gateway_ip, cidr, host_routes)

        options = {'option121': {'static_routes': static_routes}}

        body = {'name': name,
                'server_ip': server_ip,
                'dns_nameservers': dns_nameservers,
                'domain_name': dns_domain,
                'gateway_ip': gateway_ip,
                'options': options,
                'tags': tags}

        if dhcp_profile_id:
            body['dhcp_profile_id'] = dhcp_profile_id

        return body

    def build_server_config(self, network, subnet, port, tags,
                            default_dns_nameservers=None,
                            default_dns_domain=None):

        versionutils.report_deprecated_feature(
            LOG,
            'NsxLibQosNativeDhcp.build_server_config is deprecated. '
            'Please use build_server instead')

        # Prepare the configuration for a new logical DHCP server.
        server_ip = "%s/%u" % (port['fixed_ips'][0]['ip_address'],
                               netaddr.IPNetwork(subnet['cidr']).prefixlen)
        dns_nameservers = subnet['dns_nameservers']
        if not dns_nameservers or not utils.is_attr_set(dns_nameservers):
            # use the default one , or the globally configured one
            if default_dns_nameservers is not None:
                dns_nameservers = default_dns_nameservers
            else:
                dns_nameservers = self.nsxlib_config.dns_nameservers
        gateway_ip = subnet['gateway_ip']
        if not utils.is_attr_set(gateway_ip):
            gateway_ip = None
        static_routes, gateway_ip = self.build_static_routes(
            gateway_ip, subnet['cidr'], subnet['host_routes'])
        options = {'option121': {'static_routes': static_routes}}
        name = self.build_server_name(network['name'], network['id'])
        domain_name = self.build_server_domain_name(network.get('dns_domain'),
                                                    default_dns_domain)
        return {'name': name,
                'server_ip': server_ip,
                'dns_nameservers': dns_nameservers,
                'domain_name': domain_name,
                'gateway_ip': gateway_ip,
                'options': options,
                'tags': tags}
